local love = love
local lg = love.graphics

--[[local oldColor = lg.setColor
function lg.setColor(r, g, b, a)
	if type(r) == "table" then r,g,b,a = unpack(r) end
	a = a or 1
	oldColor(r*255, g*255, b*255, a*255)
end]]

local math = math
local ceil = math.ceil
local sqrt = math.sqrt

local function dist(x1, y1, x2, y2)
  return sqrt((x2 - x1) ^ 2 + (y2 - y1) ^ 2)
end

local function lerp(a, b, t)
  return a + (b - a) * t
end

local function clamp(num, min, max)
  return num < min and min or (num > max and max or num)
end

local function totalInvert(a)
  if a then return nil else return true end
end

math.randomseed(os.time())

local function newArr(w, h, f)
  local toRet = {}
  for y=1, h do
    local toIns = {}
    table.insert(toRet, toIns)
    for x=1, w do
      table.insert(toIns, type(f) == "function" and f(x, y, w, h, toRet) or f)
    end
  end
  return toRet
end

local dirs = { -- every possible direction a node can connect to. some of them have a `cross` field that specifies which connection of another node it may cross through
  {x=-1, y=0},
  {x=-1, y=-1, cross={rx=-1, ry=0}},
  {x=0, y=-1},
  {x=1, y=-1, cross={rx=1, ry=0}},
}

dirs[2].cross.dir = dirs[4]
dirs[4].cross.dir = dirs[2]

local dirsInv = {
  ["-1 0"] = dirs[1],
  ["-1 -1"] = dirs[2],
  ["0 -1"] = dirs[3],
  ["1 -1"] = dirs[4],
}

local difficulty = 0.1

local function nodeCreation(x, y, w, h, map) -- used inside of `newArr`
  local node = {x = x, y = y, count = 0, connections = {}, userCount = 0, userConnections = {}}
  for i, d in ipairs(dirs) do
    local tx, ty = x + d.x, y + d.y
    if tx >= 1 and tx <= w and ty >= 1 and ty <= h then
      node.count = node.count + 1
      node.connections[d] = true
      --print(x, y, tx, ty, map[ty])
      local other = map[ty][tx]
      other.count = other.count + 1
    end
  end
  return node
end

local function newNodeMap(w, h)
  local map = newArr(w, h, nodeCreation)
  local function removeConnection(node, d)
    node.connections[d] = nil
    node.count = node.count - 1
    local other = map[node.y + d.y][node.x + d.x]
    other.count = other.count - 1
  end
  for x=1, w do
    for y=1, h do
      local node = map[y][x]
      for d, _ in pairs(node.connections) do
        if math.random() < difficulty and node.count > 1 and map[node.y + d.y][node.x + d.x].count > 1 then
          removeConnection(node, d)
        else
          if d.cross then
            local other = map[y + d.cross.ry][x + d.cross.rx]
            if other.connections[d.cross.dir] then
              if math.random() < 0.5 then
                removeConnection(node, d)
              else
                removeConnection(other, d.cross.dir)
              end
            end
          end
        end
      end
    end
  end
  return map
end

local mapW, mapH = 5, 5

local map = newNodeMap(mapW, mapH)

local radius = 27
local margin = 43

local font = lg.newFont(radius/4*3)

local mouseOver
local selected

local lineEnd = {x = 0, y = 0}
local lineLerp = 10

local white = {1,1,1}
local correctColor = {0.5, 1, 0.5}
local wrongColor = {1, 0.5, 0.5}
local unavailColor = {0.5, 0.5, 0.5}

local function mapToScreen(x, y)
  return (x-1) * (radius*2 + margin) + radius, (y-1) * (radius*2 + margin) + radius
end

local totalW, totalH = mapToScreen(mapW, mapH)
totalW, totalH = totalW + radius, totalH + radius

local trans = {y = lg.getHeight()/2 - totalH/2}
trans.x = trans.y

local frameMar = trans.y - 50

local bgBalls = {}
for i=1, math.random(10, 20) do
  table.insert(bgBalls, {x = math.random(0, lg.getWidth()), y = math.random(0, lg.getHeight()), r = math.random(100, 200)})
end

local bgSpeed = 100

local tempCircles = {}
local animCircles = {}

local animating = false
local animDone = false

local animRadiusAdd = 70
local animSpeed = 2
local animCircleAdd = 0.2

local bgColor = {0.2,0.2,0.2}

local wiggleLine = {t = 0}
local wiggle = 10

local undoStack = {}
local redoStack = {}

local undoBtn
local redoBtn
local resetBtn
local newGameBtn
do
  local btnSize = 150
  local m = 10
  undoBtn = {x = lg.getWidth() - btnSize - m, y = m, w = btnSize, h = btnSize/2, r = 15, text = "Undo", click = function() table.insert(redoStack, table.remove(undoStack)()) end, avail = function() return #undoStack > 0 end}
  redoBtn = {x = lg.getWidth() - btnSize - m, y = undoBtn.y + undoBtn.h + m*2, w = btnSize, h = btnSize/2, r = 15, text = "Redo", click = function() table.insert(undoStack, table.remove(redoStack)()) end, avail = function() return #redoStack > 0 end}
  resetBtn = {x = lg.getWidth() - btnSize - m, y = redoBtn.y + redoBtn.h + m*5, w = btnSize, h = btnSize/2, r = 15, text = "Reset", click = function() while #undoStack > 0 do undoBtn.click() end end, avail = function() return #undoStack > 0 end}
  newGameBtn = {x = lg.getWidth() - btnSize - m, y = lg.getHeight() - btnSize/2 - m, w = btnSize, h = btnSize/2, r = 15, text = "New Game", click = function() if love.window.showMessageBox("New Game", "Start a new game?", {"Yes", "No"}) == 1 then map = newNodeMap(mapW, mapH) undoStack = {} redoStack = {} animDone = false end end, avail = function() return true end}
end

local buttons = {undoBtn, redoBtn, resetBtn, newGameBtn}

local function screenToMap(x, y)
  local tx, ty = ceil(x / (radius * 2 + margin)), ceil(y / (radius * 2 + margin))
  local ttx, tty = mapToScreen(tx, ty)
  if dist(x, y, ttx, tty) <= radius then
    return tx, ty
  end
end

local function toggleConnection(node, dir)
  node.userConnections[dir] = totalInvert(node.userConnections[dir])
  local c = node.userConnections[dir] and 1 or -1
  node.userCount = node.userCount + c
  local other = map[node.y + dir.y][node.x + dir.x]
  other.userCount = other.userCount + c
end

local function checkWin()
  for y=1, mapH do
    for x=1, mapW do
      local node = map[y][x]
      if node.userCount ~= node.count then
        return false
      end
    end
  end
  return true
end

function love.mousemoved(x, y, dx, dy)
  x, y = x - trans.x, y - trans.y
  if selected then
    local sx, sy = mapToScreen(selected.x, selected.y)
    lineEnd.x = clamp(x, sx - radius*2 - margin, sx + radius*2 + margin)
    lineEnd.y = clamp(y, sy - radius*2 - margin, sy + radius*2 + margin)
  end
  local sx,sy = selected and lineEnd.x or x, selected and lineEnd.y or y
  local tx,ty = screenToMap(sx, sy)
  if tx and tx >= 1 and tx <= mapW and ty >= 1 and ty <= mapH and map[ty][tx].count > 0 then
    mouseOver = map[ty][tx]
  else
    mouseOver = nil
  end
end

function love.mousepressed(x, y)
  x, y = x - trans.x, y - trans.y
  if mouseOver and not animating then
    if not selected then
      selected = mouseOver
      lineEnd.x, lineEnd.y = x, y
    end
  end
end

function love.mousereleased(ax, ay)
  --local x, y = ax - trans.x, ay - trans.y
  if not animating then
    local x, y = lineEnd.x, lineEnd.y
    if mouseOver then
      if selected then
        local key1 = (mouseOver.x - selected.x) .. " " .. (mouseOver.y - selected.y)
        local key2 = (selected.x - mouseOver.x) .. " " .. (selected.y - mouseOver.y)
        local changedNode, dir
        if dirsInv[key1] then
          changedNode = selected
          dir = dirsInv[key1]
        elseif dirsInv[key2] then
          changedNode = mouseOver
          dir = dirsInv[key2]
        end
        if changedNode then
          if dir.cross then
            local crossNode = map[changedNode.y + dir.cross.ry][changedNode.x + dir.cross.rx]
            if crossNode.userConnections[dir.cross.dir] then
              wiggleLine.t = 1
              wiggleLine.node = crossNode
              wiggleLine.dir = dir.cross.dir
              goto nope
            end
          end
          local other = map[changedNode.y + dir.y][changedNode.x + dir.x]
          local rfunc
          local function func()
            toggleConnection(changedNode, dir)
            return rfunc
          end
          func()
          rfunc = function()
            toggleConnection(changedNode, dir)
            return func
          end
          table.insert(undoStack, rfunc)
          redoStack = {}
          if checkWin() and not animDone then
            animDone = true
            animating = true
            for y=1, mapH do
              for x=1, mapW do
                local dx, dy = mapToScreen(x, y)
                table.insert(tempCircles, {x = dx, y = dy, t = 0})
              end
            end
            local dx, dy = mapToScreen(changedNode.x, changedNode.y)
            table.sort(tempCircles, function(a, b)
              local da, db = dist(a.x, a.y, dx, dy), dist(b.x, b.y, dx, dy)
              if da > db then
                return true
              elseif da == db and math.atan2(a.y-dy, a.x-dx) > math.atan2(b.y-dy, b.x-dx) then
                return true
              end
            end)
            table.insert(animCircles, table.remove(tempCircles))
          end
          ::nope::
        end
        selected = nil
        mouseOver = nil
      end
    else
      selected = nil
      for _, btn in ipairs(buttons) do
        if ax >= btn.x and ax <= btn.x + btn.w and ay >= btn.y and ay <= btn.y + btn.h and btn.avail() then
          btn.click()
        end
      end
    end
  end
end

function love.update(dt)
  for i,c in ipairs(animCircles) do
    local pt = c.t
    c.t = c.t + animSpeed * dt
    if animating and pt < animCircleAdd and c.t >= animCircleAdd then
      table.insert(animCircles, table.remove(tempCircles))
      bgColor[2] = correctColor[2] * 0.58
      if #tempCircles == 0 then
        animating = false
      end
    end
    if c.t >= 1 then
      table.remove(animCircles, i)
    end
  end
  for _, b in ipairs(bgBalls) do
    b.y = b.y - bgSpeed * dt
    if b.y < -b.r then
      b.r = math.random(100, 200)
      b.x = math.random(0, lg.getWidth())
      b.y = lg.getHeight() + b.r
    end
  end
  if bgColor[2] > 0.2 then
    bgColor[2] = clamp(bgColor[2] - 0.2 * dt, 0.2, 1)
  end
  if wiggleLine.t > 0 then
    wiggleLine.t = clamp(wiggleLine.t - 2 * dt, 0, 1)
  end
end

function love.draw()
  lg.setBackgroundColor(bgColor)
  lg.setColor(1,1,1,0.05)
  for _, c in ipairs(bgBalls) do
    --lg.circle("fill", c.x, c.y, c.r)
  end
  lg.push()
  lg.translate(trans.x, trans.y)
  lg.setColor(0,0,0)
  lg.rectangle("fill", -frameMar, -frameMar, totalW + frameMar*2, totalH + frameMar*2)
  lg.setColor(1, 1, 1)
  lg.setLineWidth(5)
  lg.rectangle("line", -frameMar, -frameMar, totalW + frameMar*2, totalH + frameMar*2)
  for _, c in ipairs(animCircles) do
    local r,g,b = unpack(correctColor)
    lg.setColor(r, g, b, 1 - c.t)
    lg.circle("fill", c.x, c.y, radius + c.t * animRadiusAdd)
  end
  for x=1, mapW do
    for y=1, mapH do
      local node = map[y][x]
      if node.count > 0 then
        local dx, dy = mapToScreen(x, y)
        lg.setColor(white)
        lg.setLineWidth(5)
        for d, _ in pairs(node.userConnections) do
          local ye = wiggleLine.node == node and wiggleLine.dir == d
          local cx, cy = ye and math.random(-wiggle, wiggle) * wiggleLine.t or 0, ye and math.random(-wiggle, wiggle) * wiggleLine.t or 0
          local otherdx, otherdy =  mapToScreen(x + d.x, y + d.y)
          lg.line(dx + cx, dy + cy, otherdx - cx, otherdy - cy)
        end
        if love.keyboard.isDown("f1") then
          for d, _ in pairs(node.connections) do
            local otherdx, otherdy =  mapToScreen(x + d.x, y + d.y)
            lg.line(dx, dy, otherdx, otherdy)
          end
        end
      end
    end
  end
  for x=1, mapW do
    for y=1, mapH do
      local node = map[y][x]
      if node.count > 0 then
        local dx, dy = mapToScreen(x, y)
        lg.setColor(node.userCount > node.count and wrongColor or (node.userCount == node.count and correctColor or white))
        lg.circle("fill", dx, dy, radius)
        lg.setColor(unavailColor)
        lg.setLineWidth(3)
        lg.circle("line", dx, dy, radius)
      end
    end
  end
  if mouseOver then
    lg.setLineWidth(2)
    lg.setColor(1,1,1)
    local dx, dy = mapToScreen(mouseOver.x, mouseOver.y)
    lg.circle("line", dx, dy, radius + 7)
  end
  if selected then
    lg.setLineWidth(1)
    lg.setColor(1,1,1)
    local dx, dy = mapToScreen(selected.x, selected.y)
    lg.circle("line", dx, dy, radius + 3)
    lg.setLineWidth(5)
    lg.line(dx, dy, lineEnd.x, lineEnd.y)
  end
  lg.setColor(0,0,0)
  for x=1, mapW do
    for y=1, mapH do
      local node = map[y][x]
      if node.count > 0 then
        local dx, dy = mapToScreen(x, y)
        lg.setFont(font)
        lg.printf(node.count, dx - radius, dy - font:getHeight()/2, radius*2, "center")
      end
    end
  end
  lg.pop()
  for _, btn in ipairs(buttons) do
    lg.setColor(btn.avail() and white or unavailColor)
    lg.rectangle("fill", btn.x, btn.y, btn.w, btn.h, btn.r)
    lg.setColor(0,0,0)
    lg.printf(btn.text, btn.x, btn.y + btn.h/2 - font:getHeight()/2, btn.w, "center")
  end
end